#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "X.h"
#include "window.h"
#include "defaults.h"
#include "types.h"
#include "pin.h"
#include "warning.h"
#include "list.h"
#include "mystring.h"
#include "hash.h"
#include "output.h"
#include "input.h"

char startword[] = "words{";
char startsent[] = "sentences{";

static int ctiznak(FILE *f, int backshash);
static void ctiradky(FILE *f, void (*zpracujradek)(char *, char zacr, char konr));
static char **seznamslov(char *str, char k1, char k2, int *kdekon);
static int *nactikat(char *str, int *i, char k);
static void ulozslova (char *str, char zacr, char konr);
static void ulozvetu(char *str, char zacr, char konr);

kq_word *prvulslovo = NULL;
kq_word *aktulslovo = NULL;
int pocet_word = 0;
static int pps;

static kq_list *lsents = NULL;
static void (**funkvety)(int, kq_word **);
static int dfv = 0;
static int pfv;

kq_word *add_word (char kontext)
{
  kq_word *vysledek;

  if(win){
    vysledek = aktulslovo;
    aktulslovo = aktulslovo->next;
  }
  else{
    vysledek = (kq_word *)malloc(sizeof(kq_word));
    vysledek->next = NULL;
    if(!prvulslovo) aktulslovo = prvulslovo = vysledek;
    else aktulslovo->next = vysledek;
    aktulslovo = vysledek;
    vysledek->cislo = ++pocet_word;
    vysledek->nazev = NULL;
    vysledek->informace = pinnull;
    vysledek->typ = STDWORD;
  }
  vysledek->kontext = kontext;

  return vysledek;
}

void add_sent (void (*f)(int, kq_word **))
{
  if(win) return;
  addPdoL(f, &lsents);
  dfv++;
}

void finishws ()
{
  funkvety = (void (**)(int, kq_word **))LdoPA(lsents);
  smazlist(&lsents);
}

struct veta *jenkat = NULL;


void warneznam (char *slovo)
{
  warning("nezname slovo: \"%s\"", slovo);
}

void popwarneznam (char *slovo)
{
  popupwarning("_unknown word", slovo);
}

static int ctiznak (FILE *f, int backslash)
{
  char ch;

  ch=fgetc(f);
  if(backslash && ch == '\\'){
    while((ch = ctiznak(f, 0)) == ' ' || ch == '\t');
    if(ch == '\n') return ' ';
  }
  else if(ch == '%'){
    while((ch = fgetc(f)) != '\n' && ch != EOF);
  }
  return ch;
}

static void ctiradky (FILE *f, void (*zpracujradek)(char *, char zacr, char konr))
{
  char *str, ch, zacr, konr;
  kq_string radek;

  zacr = 1;
  initstr(&radek);
  for(;;){
    while((ch=ctiznak(f,1)) == '\n' || ch == ' ' || ch == '\t');
    if(ch == EOF || ch == '}'){
      if(ch == EOF) warning("predcasny konec souboru");
      return;
    }
    for(;;){
      if(ch == '\n' || ch == EOF || ch == '}' || ch == ';'){
	str = zestringuj(&radek);
	if(ch == ';') konr = 0;
	else konr = 1;

	zpracujradek(str, zacr, konr);

	smazstr(&radek);
	free(str);
	if(ch == ';') zacr = 0;
	else zacr = 1;

	if(ch == EOF || ch == '}'){
	  if(ch == EOF) warning("predcasny konec souboru");
	  return;
	}
	if(ch != ';') break;
      }
      else zapisznak(ch, &radek);

      ch=ctiznak(f,1);
    }
  }
}

static int poslkat;

void nactiinput (char *jazyk)
{
  char *fullname;
  FILE *f;
  int ch, w, s;

  fullname = joinstrings(DATADIR, LANGDIR, jazyk, INPUTSUFFIX, NULL);
  if((f = fopen (fullname, "r")) == NULL){
    warning("Nepovedlo se otevrit soubor %s", fullname);
    free(fullname);
    return;
  }

  pps = pfv = 0;
  aktulslovo = prvulslovo;

  w=s=0;

  poslkat = 2;

  for(;;){
    ch = fgetc(f);
    if(ch == EOF) break;
    else if(ch == '\n') w=s=0;
    else{
      if(w >= 0 && ch == startword[w]){
	w++;
	if(!startword[w]){
	  w = -1;
	  s = -1;
	  ctiradky(f, &ulozslova);
	}
      }
      else w = -1;

      if(s >= 0 && ch == startsent[s]){
	s++;
	if(!startsent[s]){
	  w = -1;
	  s = -1;
	  ctiradky(f, ulozvetu);
	}
      }
      else s = -1;
    }
  }

  if(pps != pocet_word) warning("spatny pocet slov, zrojak: %d, soubor: %d", pocet_word, pps);
  if(pfv != dfv) warning("spatny pocet vet, zrojak: %d, soubor: %d", dfv, pfv);

  free(fullname);
  fclose(f);
}

static int katnumber (char *katnazev)
{
  kq_pin *p, p2;

  p = hash_find(WORDCAT, katnazev);
  if(!p || p->typ != PININT){
    if(p) warning("kategorie ma v hashi spatny typ");
    p2 = pinint(poslkat++);
    p = &p2;
    hash_add(WORDCAT, katnazev, p);
  }
  return p->u.i;
}

static int *nactikat (char *str, int *i, char k)
{
  int j;
  int *vysledek;
  char *pomstr;
  kq_list *l;

  l = NULL;

  for(j = ++*i;; j++)
    if(str[j] == ',' || str[j] == k || str[j] == 0){
      pomstr = dupnstr(&str[*i], j-*i);
      upravstring(pomstr);

      if(pomstr[0]) addIdoL(katnumber(pomstr), &l);
      free(pomstr);

      if(!str[j]){
	warning("chybi zaviraci zavorka '%c'", k);
	*i=j;
	break;
      }

      *i=j+1;

      if(str[j] == k) break;
    }

  if(!l) return NULL;

  vysledek = LdoIA(l);
  smazlist(&l);
  return vysledek;
}

static char **seznamslov (char *str, char k1, char k2, int *kdekon)
{
  int i,j,k, s,t, delka;
  char ch;
  kq_list *l;

  char **s1, **s2, *vysls, **vysledek;

  l = NULL;

  if(k1 == ']' && k2 ==0) addPdoL(strdup(""), &l); // hranate zavorky jsou nepovinne

  for(s=i=*kdekon;;){
    if(str[i] == 0 || str[i] == '=' || str[i] == k1 || str[i] == k2){
      addPdoL(dupnstr(&str[s], i-s), &l);
      break;
    }
    else if(str[i] == ']'){
      warning("nezparovana zavorka ']'");
      i++;
    }
    else if(str[i] == ')'){
      warning("nezparovana zavorka ')'");
      i++;
    }
    else if(str[i] == '[' || str[i] == '('){
      t = i-s;

      if(str[i] == '(') ch = ')';
      else ch = ']';
      i++;
      s1 = seznamslov(str, ch, 0, &i);
      i++;
      s2 = seznamslov(str, k1, ',', &i);

      delka = t+1;
      for(j=0; s1[j]; j++){
	delka += strlen(s1[j]);
	for(k=0; s2[k]; k++){
	  delka += strlen(s2[k]);

	  vysls = malloc(sizeof(char)*delka);
	  strncpy(vysls, &str[s], t);
	  strcpy(&vysls[t], s1[j]);
	  strcpy(&vysls[t+strlen(s1[j])], s2[k]);
	  addPdoL(vysls, &l);

	  delka -= strlen(s2[k]);
	}
	delka -= strlen(s1[j]);
      }

      for(j=0; s1[j]; j++) free(s1[j]);
      for(k=0; s2[k]; k++) free(s2[k]);
      free(s1);
      free(s2);

      if(str[i] == k1 || str[i] == k2 || str[i] == '=' || str[i] == 0) break;
      s = ++i;
    }
    else if(str[i] == ','){
      addPdoL(dupnstr(&str[s], i-s), &l);
      s = ++i;
    }
    else i++;
  }

  if((str[i] == '=' || str[i] == 0) && k1){
    warning("Predcasne ukonceni slova (chybi uzaviraci zavorka)");
    i--;
  }
  *kdekon = i;
  vysledek = (char **)LdoPA(l);
  smazlist(&l);

  return vysledek;
}

void wordtohash (struct slovo *w, char *str)
{
  kq_pin p;
  int i;

  p = pinpointer(w);

  hash_add(WORD, str, &p);

  for(i=strlen(str)-1; i; i--)
    if(str[i] == ' '){
      str[i] = 0;
      hash_add(WORDPART, str, &pinnull);
      str[i] = ' ';
    }
}

static kq_word *tus;

static void ulozslova (char *str, char zacr, char konr)
{
  char **slova;
  struct slovo *slovo;
  kq_list *lpompol;
  int *kateg;
  int i, j;

  for(i=0; str[i] == ' ' || str[i] == '\t' ; i++);

  if(str[i] == '<'){
    kateg = nactikat(str, &i, '>');
    if(zacr){
      if(pps < pocet_word){
	tus = aktulslovo;
	aktulslovo = aktulslovo->next;
      }
      else tus = NULL;
      pps++;
    }
  }
  else{
    kateg = NULL;
    if(zacr) tus = NULL;
  }

  slova = seznamslov(str, 0, 0, &i);
  j = i;

  // Zalolozen struktury

  slovo = (struct slovo *)malloc(sizeof(struct slovo));
  slovo->ulozeno = NULL;
  slovo->kateg = kateg;
  slovo->pocet = 0;
  slovo->makro = 0;
  slovo->u.vety = NULL;
  slovo->ulozeno = tus;
  if(zacr && tus && tus->nazev){
    free(tus->nazev);
    tus->nazev = NULL;
  }

 
  // Pridani do hashe:

  for(i=0; slova[i]; i++){
    upravstring(slova[i]);
    if(!slova[i][0]){
      free(slova[i]);
      continue;
    }

    if(tus && !tus->nazev) tus->nazev = strdup(slova[i]);

    slovo->pocet++;
    wordtohash(slovo, slova[i]);

    free(slova[i]);
  }
  free(slova);

  if(slovo->pocet == 0){
    if(kateg){
      warning("prazdne slovo ma kategorie");
      free(kateg);
    }
    free(slovo);
  }
  if(konr && tus && !tus->nazev) warning("ukladane slovo nema nazev");

  if(str[j] == '='){
    if(!slovo){
      warning("makro slovo je prazdne");
      return;
    }
    if(kateg || tus)
      warning("makroslovo je prekombinovane");
    lpompol = rozeberstr(&str[j], &warneznam);
    expandvetu(&lpompol);
    slovo->makro = 1;
    slovo->u.nahradit = lpompol;
  }
}

static void ulozvetu (char *str, char zacr, char konr)
{
  int i,j,k, pocetp, pocetc, jeparam;
  char ch;
  struct veta *v;
  kq_list *l;
  struct slovo **pompol, *s;
  int *kateg;

  struct vetnyclen vc[26];
  char nahrstr[28];

  pocetp = pocetc = 0;
  jeparam = 0;

// definice vety:

  for(i=0;;){
    for(j=i; str[j] != '<' && str[j] != '[' && str[j] != ','
	  && str[j] != '(' && str[j] != ')' && str[j] != 0; j++);
    ch = str[j];
    str[j] = 0;

    l = rozeberstr(&str[i], &warneznam);
    pompol = (struct slovo **)LdoPA(l);
    smazlist(&l);

    for(k=0; pompol[k]; k++){
      if(pocetc < 26){
	vc[pocetc].jekat = 0;
	vc[pocetc].u.presneslovo = pompol[k];
	if(jeparam){
	  vc[pocetc].mistoparam = pocetp;
	  if(vc[pocetc].u.presneslovo->makro)
	    pocetp += delkaL(vc[pocetc].u.presneslovo->u.nahradit);
	  else pocetp++;
	}
	else vc[pocetc].mistoparam = -1;
      }
      pocetc++;
    }
    free(pompol);

    str[j] = ch;

    if(ch == 0 || ch == ',') break;

    if(ch == '<' || ch == '['){
      if(ch == '<') kateg = nactikat(str, &j, '>');
      else kateg = nactikat(str, &j, ']');
      if(pocetc < 26){
	vc[pocetc].jekat = 1;
	vc[pocetc].u.kateg = kateg;
	if(ch == '<' || jeparam) vc[pocetc].mistoparam = pocetp;
	else vc[pocetc].mistoparam = -1;
      }
      else free(kateg);
      pocetc++;
      if(ch == '<' || jeparam) pocetp++;
      i=j;
    }
    else if(ch == '('){
      if(jeparam) warning("vlozena zavorka '(' v definici vety");
      jeparam = 1;
      i=j+1;
    }
    else if(ch == ')'){
      if(!jeparam) warning("nezparovana zavorka ')' v definici vety");
      jeparam = 0;
      i=j+1;
    }
    else warning("neco shnileho");
  }

// poradi vety:

  if(!str[j]){
    str = nahrstr;
    str[0] = ',';
    for(j=0; j-1<pocetc && 'a'-1 + j <= 'z'; j++) str[j] = 'a'-1 + j;
    str[j] = 0;
    j=0;
  }

  while(str[j]){
    v = (struct veta *)malloc(sizeof(struct veta));
    v->pocetclenu = 0;
    v->pocetparam = pocetp;
    if(pfv < dfv) v->funkce = funkvety[pfv];

    // spocitam cleny:

    j++;
    i=j;
    while(str[j] && str[j] != ','){
      str[j] = tolower(str[j]);
      if(str[j] != ' '){
	if(str[j] < 'a' || str[j] > 'z'){
	  warning("v urceni poradi vety je jiny znak nez pismeno anglicke abecedy");
	}
	else if(str[j]-'a' >= pocetc){
	  warning("v urceni poradi vety je prilis vysoke pismenko");
	}
	else{ 
	  if(!vc[str[j]-'a'].jekat &&
	     vc[str[j]-'a'].u.presneslovo->makro)
	  v->pocetclenu += delkaL(vc[str[j]-'a'].u.presneslovo->u.nahradit);
	  else v->pocetclenu++;
	}
      }
      j++;
    }

    v->cleny = (struct vetnyclen *)malloc((v->pocetclenu)*sizeof(struct vetnyclen));

    // ulozim cleny:

    j=i;
    k=0;
    while(str[j] && str[j] != ','){
      if(str[j] >= 'a' && str[j] <= 'z' && str[j]-'a' < pocetc){
	if(!vc[str[j]-'a'].jekat){
	  if(vc[str[j]-'a'].u.presneslovo->makro){
	    l = vc[str[j]-'a'].u.presneslovo->u.nahradit;
	    i = delkaL(l);
	    jeparam = vc[str[j]-'a'].mistoparam;
	    if(jeparam >= 0) jeparam += i;
	    k += i;
	    i = k;
	    while(l){
	      k--;
	      v->cleny[k].jekat = vc[str[j]-'a'].jekat;
	      if(jeparam >= 0)
		jeparam--;
	      v->cleny[k].mistoparam = jeparam;
	      v->cleny[k].u.presneslovo = l->u.p;
	      l = l->next;
	    }
	    k=i;
	  }
	  else{
	    v->cleny[k].jekat = vc[str[j]-'a'].jekat;
	    v->cleny[k].mistoparam = vc[str[j]-'a'].mistoparam;
	    v->cleny[k].u.presneslovo = vc[str[j]-'a'].u.presneslovo;
	    k++;
	  }
	}
	else{
	  v->cleny[k].jekat = vc[str[j]-'a'].jekat;
	  v->cleny[k].mistoparam = vc[str[j]-'a'].mistoparam;
	  if(vc[str[j]-'a'].u.kateg){
	    for(i=0; vc[str[j]-'a'].u.kateg[i]; i++);

	    v->cleny[k].u.kateg = (int *)malloc((i+1)*sizeof(int));

	    for(i=0; vc[str[j]-'a'].u.kateg[i]; i++)
	      v->cleny[k].u.kateg[i] = vc[str[j]-'a'].u.kateg[i];
	    v->cleny[k].u.kateg[i] = 0;
	  }
	  else{
	    v->cleny[k].u.kateg = NULL;
	  }
	  k++;
	}
      }
      j++;
    }
    if(k != v->pocetclenu)
      warning("ulozvetu: spatne spocteny pocet slov, k=%d, v->pocetclenu = %d",
	      k, v->pocetclenu);

    for(i=0; i<v->pocetclenu && v->cleny[i].jekat; i++);

    if(i == v->pocetclenu){
      v->next = jenkat;
      jenkat = v;
    }
    else{
      s = v->cleny[i].u.presneslovo;
      if(s->makro) warning("ulozvetu: prvni slovo ma spatny typ");

      v->next = s->u.vety;
      s->u.vety = v;
    }
  }
  for(i=0; i<pocetc; i++)
    if(vc[i].jekat == 1) free(vc[i].u.kateg);

  if(konr) pfv++;
}

//: rozeberstr

kq_list *rozeberstr (char *str, void (*neznam)())
{
  char *pomstr, ch;
  int i,j,k;
  kq_pin *p, *pp;
  kq_list *vysledek;

  pomstr = strdup(str);
  upravstring(pomstr);

  if(!pomstr[0]){
    free(pomstr);
    return NULL;
  }

  p = NULL;
  vysledek = NULL;

  for(i=j=0;; i++)
    if(pomstr[i] == ' ' || pomstr[i] == 0){
      ch = pomstr[i];
      pomstr[i] = 0;

      if((pp = hash_find(WORD, &pomstr[j]))) k = i;
      if(pp) p = pp;

      if(!hash_find(WORDPART, &pomstr[j]) || str[i] == 0){
	if(!p){
	  neznam(&pomstr[j]);
	  smazlist(&vysledek);
	  free(pomstr);
	  return NULL;
	}
	if(p->typ != PINPOINTER) warning("rozeberstr: slovo v hashi neni ukazatel");

	pomstr[i] = ch;

	addPdoL(p->u.p, &vysledek);

	p = NULL;
	if(k == i && ch == 0) break;
	i=j=k+1;
      }
      else pomstr[i] = ch;
    }

  free(pomstr);
  return vysledek;
}

//: expandvetu

void expandvetu (kq_list **v)
{
  kq_list *pom, *pomo;

  while(*v){
    if(((struct slovo *)((*v)->u.p))->makro){
      pom = ((struct slovo *)((*v)->u.p))->u.nahradit;
      if(pom){
	(*v)->u.p = pom->u.p;

	for(pom = pom->next; pom; pom = pom->next){
	  pomo = (kq_list *)malloc(sizeof(kq_list));
	  pomo->u.p = pom->u.p;
	  pomo->next = (*v)->next;
	  (*v)->next = pomo;
	  v = &(*v)->next;
	}
      }
      else{
	pomo = *v;
	*v = pomo->next;
	free(pomo);
      }
    }
    else v = &(*v)->next;
  }
}

//: provedvetu

char provedvetu (struct slovo **slova)
{
  int i, j, k, l, d;
  struct veta *v;
  kq_word **parametry;

  for(d=0; slova[d]; d++)
    if(slova[d]->ulozeno && slova[d]->ulozeno->kontext <= 0){
      popupwarning("_bad context", slova[d]->ulozeno->nazev);
      return 0;
    }

  for(i=0;;i++){
    if(slova[i]){
      if(slova[i]->makro) warning("provedvetu: spatny typ %d. slova", i);
      v = slova[i]->u.vety;
    }
    else v = jenkat;

    for(;v ; v = v->next){
      if(v->pocetclenu == d && (!slova[i] ||
				v->cleny[i].u.presneslovo == slova[i])){
	for(j=0; j<d; j++){
	  if(v->cleny[j].jekat){
	    if(!slova[j]->kateg || !v->cleny[j].u.kateg) break;
	    for(k=0; slova[j]->kateg[k]; k++){
	      for(l=0; v->cleny[j].u.kateg[l]; l++)
		if(v->cleny[j].u.kateg[l] == slova[j]->kateg[k]) break;
	      if(v->cleny[j].u.kateg[l] == slova[j]->kateg[k]) break;
	    }
	    if(!slova[j]->kateg[k]) break;
	  }
	  else{
	    if(v->cleny[j].u.presneslovo != slova[j]) break;
	  }
	}
	if(j == d){
	  if(v->pocetparam && v->funkce){
	    parametry = (kq_word **)malloc(v->pocetparam*sizeof(kq_word *));
	    for(j = 0; j<v->pocetparam; j++) parametry[j] = NULL;
	    for(j = 0; j<d; j++)
	      if(v->cleny[j].mistoparam >= 0){
		if(slova[j]->ulozeno) parametry[v->cleny[j].mistoparam] = slova[j]->ulozeno;
		else warning("provedvetu: neulozene slovo chce do parametru %d",
			     v->cleny[j].mistoparam);
	      }
	  }
	  else parametry = NULL;

	  if(v->funkce) v->funkce(v->pocetparam, parametry);
	  if(parametry) free(parametry);

	  return 1;
	}
      }
    }
    if(!slova[i]){
      popupwarning("_unknown sentence", "#");
      return 0;
    }
  }
}

//: provedstr

char provedstr(char *str)
{
  char vysledek;
  kq_list *l;
  struct slovo **slova;

  l = rozeberstr(str, &popwarneznam);
  if(!l){
    smazlist(&l);
    return 0;
  }
  expandvetu(&l);
  slova = (struct slovo **)LdoPA(l);

  vysledek = provedvetu(slova);
  free(slova);
  return vysledek;
}
