#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include "defaults.h"
#include "warning.h"
#include "types.h"
#include "pin.h"
#include "list.h"
#include "mystring.h"
#include "hash.h"
#include "X.h"
#include "popup.h"
#include "input.h"
#include "output.h"
#include "languages.h"

kq_word *aktulang = NULL;
kq_word *prvlang = NULL;
kq_word *poslang = NULL;

void nactijazyky()
{
  char *fullname, *str;
  FILE *f;
  kq_string nekos;
  char ch;
  int i;

  fullname = joinstrings(DATADIR, LANGDIR, LANGFILE, NULL);
  if((f = fopen (fullname, "r")) == NULL){
    warning("Nepovedlo se otevrit soubor %s", fullname);
    free(fullname);
    return;
  }

  initstr(&nekos);
  for(;;){
    ch = fgetc(f);
    if(ch == EOF) break;

    while(ch == ' ' || ch == '\t') ch = fgetc(f);
    while(ch != '=' && ch != '\n' && ch != EOF){
      zapisznak(ch, &nekos);
      ch = fgetc(f);
    }
    str = zestringuj(&nekos);
    for(i = nekos.delka-1; i >= 0 && (str[i] == ' ' || str[i] == '\t'); i--);
    str[i+1] = 0;
    smazstr(&nekos);
    if(!str[0] && ch != '='){
      free(str);
      continue;
    }

    if(poslang){
      poslang->next = (kq_word *)malloc(sizeof(kq_word));
      poslang = poslang->next;
    }
    else prvlang = poslang = (kq_word *)malloc(sizeof(kq_word));

    poslang->informace = pinpointer(str);
    poslang->typ = LANGWORD;
    poslang->kontext = 1;
    poslang->next = NULL;

    if(ch == '='){
      ch = fgetc(f);
      while(ch != '\n' && ch != EOF){
	zapisznak(ch, &nekos);
	ch = fgetc(f);
      }
      poslang->nazev = zestringuj(&nekos);
      upravstring(poslang->nazev);
      if(!poslang->nazev[0]){
	free(poslang->nazev);
	poslang->nazev = NULL;
      }
      smazstr(&nekos);
    }
    else poslang->nazev = str;
  }
  fclose(f);
  free(fullname);
}

static void jazslova()
{
  kq_word *pom;
  struct slovo *slovo;
  struct veta *puvveta;
  kq_pin p;

  p = pinint(1);
  hash_add(WORDCAT, LANGCAT, &p);

  for(pom=prvlang; pom; pom = pom->next){
    if(pom->nazev){
      slovo = (struct slovo *)malloc(sizeof(struct slovo));
      slovo->pocet = 1;
      slovo->makro = 0;
      slovo->kateg = (int *)malloc(2*sizeof(int));
      slovo->kateg[0] = 1;
      slovo->kateg[1] = 0;
      slovo->u.vety = NULL;
      slovo->ulozeno = pom;

      wordtohash(slovo, pom->nazev);
    }
  }

  puvveta = jenkat;
  jenkat = (struct veta *)malloc(sizeof(struct veta));
  jenkat->funkce = prepnilang;
  jenkat->pocetclenu = 1;
  jenkat->pocetparam = 1;
  jenkat->cleny = (struct vetnyclen *)malloc(sizeof(struct vetnyclen));
  jenkat->cleny->jekat = 1;
  jenkat->cleny->u.kateg = (int *)malloc(sizeof(int));
  jenkat->cleny->u.kateg[0] = 1;
  jenkat->cleny->u.kateg[1] = 0;
  jenkat->cleny->mistoparam = 0;
  jenkat->next = puvveta;
}

static void freesents(struct veta *v)
{
  int i;
  struct veta *v2;

  for(; v; v = v2){
    for(i=0; i < v->pocetclenu; i++)
      if(v->cleny[i].jekat) free(v->cleny[i].u.kateg);
    free(v->cleny);
    v2=v->next;
    free(v);
  }
}

static void freeword(struct slovo *w)
{
  w->pocet--;
  if(w->pocet) return;

  if(w->makro) smazlist(&w->u.nahradit);
  else freesents(w->u.vety);
  if(w->kateg) free(w->kateg);
  free(w);
}

static void zrusjazyk()
{
  int i;
  struct hash **pom, *pomo, *pomoc;

  for(i=0; i<HASHSIZE; i++)
    for(pom = &hashtable[i]; *pom; pom = &pomo->next){
      for(pomo = *pom; pomo && (pomo->typ == WORD || pomo->typ == WORDPART ||
				pomo->typ == WORDCAT || pomo->typ == OUTPUT);
				pomo = pomo->next){
	pomoc = pomo->next;
	if(pomo->typ == OUTPUT) freeoutputs(pomo->informace.u.p);
	if(pomo->typ == WORD) freeword(pomo->informace.u.p);
	else free(pomo);
      }
      *pom = pomo;
      if(!pomo) break;
    }

  freesents(jenkat);
  jenkat = NULL;
}

void startlang()
{
  jazslova();
  if(aktulang){
    nactiinput(aktulang->informace.u.p);
    nactioutput(aktulang->informace.u.p);
  }
}

void zmenlang(kq_word *jazyk)
{
  aktulang = jazyk;
  zrusjazyk();
  jazslova();
  nactiinput(jazyk->informace.u.p);
  nactioutput(jazyk->informace.u.p);
}

void prepnilang(int pocet, kq_word **jazslovo)
{
  char *jaznavrat, *jazprichod, *popstr;

  if(pocet != 1){
    warning("prepnilang: spatny pocet slov");
    return;
  }

  if(jazslovo[0]->typ != LANGWORD){
    warning("prepnilang: jazslovo[0]->typ != LANGWORD");
    return;
  }

  if(jazslovo[0] == aktulang){
    popupoutput("_same language", NULL, 0.9, 0.9, 0.5, 0.5);
    return;
  }

  jaznavrat = outputstring("_return to language", NULL);
  if(jaznavrat) jaznavrat = strdup(jaznavrat);

  zmenlang(jazslovo[0]);
  
  jazprichod = outputstring("_language start", NULL);
  if(jazprichod){
    if(jaznavrat) popstr = joinstrings(jazprichod, "\n", jaznavrat, NULL);
    else popstr = jazprichod;
    popupstring(0.9, 0.9, 0.5, 0.5, popstr);
    if(popstr != jazprichod) free(popstr);
  }
  else popupoutput("_language start", NULL, 0.9, 0.9, 0.5, 0.5);

  if(jaznavrat) free(jaznavrat);
}

void zobrazjazyky()
{
  char **jazlist;
  kq_word *pom;
  int i, aktu;

  i=1;
  for(pom = prvlang; pom; pom = pom->next){
    if(!pom->nazev){
      warning("neznamy nazev jazyku \"%s\"", pom->informace.u.p);
      return;
    }
    if(pom == aktulang) aktu = i-1;
    i++;
  }

  jazlist = (char **)malloc(i*sizeof(char *));

  i=0;
  for(pom = prvlang; pom; pom = pom->next){
    jazlist[i] = pom->nazev;
    i++;
  }
  jazlist[i] = NULL;

  i = popupselect(0.9, 0.9, 0.5, 0.5,
		  outputstring("_languages", NULL),
		  jazlist, aktu, NULL);
  free(jazlist);

  if(i >= 0 && i != aktu){
    for(pom = prvlang; i; i--) pom = pom->next;
    zmenlang(pom);
  }
}
